"""
Contains class representation of executable code block.
"""
from prompt_toolkit import prompt
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments_cls
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import get_lexer_by_name
from rundoc import BadEnv, CodeFailed, BadInterpreter
import logging
import os
import select
import subprocess
import sys
import time

class DocBlock(object):
    """Single multi-line code block executed as a script.

    Attributes:
        interpreter (str): Interpreter used to run the code.
        code (str): Base code loaded during initialization.
        process (subprocess.Popen): Process object running the interpreter.
        runs (list): List of dictinaries, each containing the following:
            'user_code': User modified version of the code (will be used
                instead of main code unless it's set to None or empty string).
            'output': Full output of executed code block.
            'retcode': exit code of the code block executed
    """
    def __init__(self, code, interpreter, darkbg=True, tags=""):
        if darkbg:
            from pygments.styles.native import NativeStyle as HighlightStyle
            self.HighlightStyle = HighlightStyle
        else:
            from pygments.styles.manni import ManniStyle as HighlightStyle
            self.HighlightStyle = HighlightStyle
        self.interpreter = interpreter
        self.code = code
        self.tags = tags
        self.process = None
        self.runs = []  # elements inside are like:
                        #   {
                        #       'user_code':'',
                        #       'output':'',
                        #       'retcode':None,
                        #       'time_start': None,
                        #       'time_stop': None
                        #   }
        interpreter_exists = subprocess.call(
            ['bash', '-c', 'command -v {} 2>&1>/dev/null'.format(interpreter)])
        if interpreter_exists != 0:
            raise BadInterpreter("Bad interpreter: '{}'".format(interpreter))

    @property
    def last_run(self):
        if len(self.runs):
            return self.runs[-1]
        else:
            return None

    def get_lexer_class(self):
        lexer_class = None
        try:
            # try because lexer may not exist for current interpreter
            return get_lexer_by_name(self.interpreter).__class__
        except:
            # no lexer, return plain text
            return None

    def __str__(self):
        lexer_class = self.get_lexer_class()
        code = ''
        if self.last_run:
            code = self.last_run['user_code'].strip()
        else:
            code = self.code
        if lexer_class:
            return highlight(
                code,
                lexer_class(),
                Terminal256Formatter(style=self.HighlightStyle)
                )
        return code

    def get_dict(self):
        return {
            'interpreter': self.interpreter,
            'code': self.code,
            'tags': self.tags,
            'runs': self.runs,
        }

    def prompt_user(self, prompt_text='» '):
        self.last_run['user_code'] = prompt(
            prompt_text,
            default = self.code,
            lexer = PygmentsLexer(self.get_lexer_class()),
            style = style_from_pygments_cls(self.HighlightStyle)
            )

    def print_output(self, final=False):
        """Read both stdout and stderr, populate them in the variable and print.

        Args:
            final (bool): Used to collect final bytes after the process exists.
        """
        encoding = sys.stdout.encoding
        if final:
            line = self.process.stderr.read().decode(encoding)
            self.last_run['output'] += line
            sys.stderr.write(line)
            line = self.process.stdout.read().decode(encoding)
            self.last_run['output'] += line
            sys.stdout.write(line)
        else:
            assert self.process
            reads = [self.process.stdout.fileno(), self.process.stderr.fileno()]
            ret = select.select(reads, [], [])
            line = ""
            for fd in ret[0]:
                if fd == self.process.stderr.fileno():
                    line = self.process.stderr.readline().decode(encoding)
                    self.last_run['output'] += line
                    sys.stderr.write(line)
                if fd == self.process.stdout.fileno():
                    line = self.process.stdout.readline().decode(encoding)
                    self.last_run['output'] += line
                    sys.stdout.write(line)
            return len(line) > 0

    def is_running(self):
        return self.process and self.process.poll() is None

    def run(self, prompt=True):
        if not self.process:
            self.runs.append(
                {
                    'user_code': '',
                    'output': '',
                    'retcode': None,
                    'time_start': None,
                    'time_stop': None,
                }
            )
            if prompt:
                self.prompt_user()
            else:
                self.last_run['user_code'] = self.code
            code = self.last_run['user_code'].strip()
            self.last_run['time_start'] = time.time()
            self.process = subprocess.Popen(
                [self.interpreter, '-c', code],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=False,
                )
        while self.is_running():
            self.print_output()
        self.print_output(final=True)
        self.last_run['time_stop'] = time.time()
        self.last_run['retcode'] = self.process.poll()
        self.process = None

    def kill(self):
        if self.process:
             self.process.kill()





"""
Main module for rundoc command line utility.
"""
from rundoc import BadEnv, CodeFailed
from rundoc import parsers
import argcomplete
import argparse
import logging
import rundoc
import sys

def __parse_args():
    """Parse command line arguments and return an argparse object."""
    parser = argparse.ArgumentParser(
        prog = "rundoc",
        description="Run code from markdown code blocks in controlled manner",
        )
    parser.add_argument(
        "-v", "--version", action="store_true",
        help="Show version info and exit."
        )
    parser.add_argument(
        "-d", "--debug", action="store_true",
        help="Enable debug mode with output of each action in the log."
        )
    subparsers = parser.add_subparsers(
        description="(use each command with -h for more help)",
        dest="cmd",
        )

    parser_run = subparsers.add_parser(
        "run",
        description='''
            Extract all code blocks from markdown file and run them as script.
            First tag in every code block (highlighting tag) will be used as
            name of the interpreter for that code block. Code blocks can be
            included and excluded using '-t', '-T' and '-N' tags.'''
        )
    parser_run.add_argument(
        "mkd_file", type=str, action="store",
        help="Markdown file path."
        )
    parser_run.add_argument(
        "-t", "--tags", action="store",
        help='''Hash (#) separated list of tags (e.g. -t bash#proj-2#test_3).
                Filter out all code blocks that are missing all of the listed
                tags. (Works like OR logic).
                '''
        )
    parser_run.add_argument(
        "-T", "--must-have-tags", action="store",
        help='''Hash (#) separated list of tags (e.g. -t bash#proj-2#test_3).
                Filter out all code blocks that are missing at least one of the
                listed tags. (Works like AND logic).'''
        )
    parser_run.add_argument(
        "-N", "--must-not-have-tags", action="store",
        help='''Hash (#) separated list of tags (e.g. -t bash#proj-2#test_3).
                Filter out all code blocks that contain at least one of the
                listed tags. (Works like AND NOT logic).'''
        )
    parser_run.add_argument(
        "-i", "--inherit-env", action="store_true",
        help='''Override default env variable values with values from existing
                environment. Useful when you export variables before running
                rundoc and expect them to override default values that are set
                in the docs.'''
        )
    parser_run.add_argument(
        "-p", "--pause", type=int, action="store",
        help='''Used in combination with -y. Pause N seconds before each code
                block. Defaults to 0.'''
        )
    parser_run.add_argument(
        "-s", "--step", type=int, action="store",
        help="Start at step STEP. Defaults to 1."
        )
    parser_run.add_argument(
        "-o", "--output", type=str, action="store",
        help='''Output file. All codes, modifications and output of execution
                will be saved here. This file can be later used as an argument
                to 'replay' command which will execute all steps without any
                user interaction.'''
        )
    parser_run.add_argument(
        "--light", action="store_true",
        help="Light terminal background."
        )
    parser_run.add_argument(
        "-y", "--yes", action="store_true",
        help="Confirm all steps with no user interaction."
        )
    parser_run.add_argument(
        "-r", "--retry", type=float, action="store",
        help='''Number of retries for failed code blocks when using -y option.
                Defaults to 0.'''
        )
    parser_run.add_argument(
        "-P", "--retry-pause", type=float, action="store",
        help='''Pause N seconds before retrying. Actual pause will be max value
                of --pause or --retry-pause. Defaults to 1.'''
        )
    parser_run.set_defaults(
        tags="",
        must_have_tags="",
        must_not_have_tags="",
        inherit_env=False,
        pause=0,
        step=1,
        retry=0,
        retry_pause=1,
        output=None,
        )

    parser_replay = subparsers.add_parser(
        "replay",
        description="Run codes saved as json output by 'run' command."
        )
    parser_replay.add_argument(
        "saved_output", type=str, action="store",
        help="Path of saved output file from 'run' or 'replay' command."
        )
    parser_replay.add_argument(
        "-p", "--pause", type=int, action="store",
        help="Pause N seconds before each code block. Defaults to 0."
        )
    parser_replay.add_argument(
        "-s", "--step", type=int, action="store",
        help="Start at step STEP. Defaults to 1."
        )
    parser_replay.add_argument(
        "-o", "--output", type=str, action="store",
        help='''Output file. All codes, modifications and output of execution
                will be saved here. This file can be later used as an input
                argument to 'replay' command.'''
        )
    parser_replay.add_argument(
        "--light", action="store_true",
        help="Light terminal background."
        )
    parser_replay.add_argument(
        "-r", "--retry", type=float, action="store",
        help="Number of retries for failed code blocks. Defaults to 0."
        )
    parser_replay.add_argument(
        "-P", "--retry-pause", type=float, action="store",
        help='''Pause N seconds before retrying. Actual pause will be max value
                of --pause or --retry-pause. Defaults to 1.'''
        )
    parser_replay.set_defaults(
        pause=0,
        step=1,
        retry=0,
        retry_pause=1,
        output=None,
        )

    parser_list_tags = subparsers.add_parser(
        "list-tags",
        description='''List all unique tags that appear in the markdown file
                    and number of occurrences.'''
        )
    parser_list_tags.add_argument(
        "mkd_file", type=str, action="store",
        help="Markdown file path."
        )

    parser_list_blocks = subparsers.add_parser(
        "list-blocks",
        description='''List all blocks that would be executed with selected
            tags and parameters.'''
        )
    parser_list_blocks.add_argument(
        "mkd_file", type=str, action="store",
        help="Markdown file path."
        )
    parser_list_blocks.add_argument(
        "-t", "--tags", action="store",
        help='''Hash (#) separated list of tags (e.g. -t bash#proj-2#test_3).
                Filter out all code blocks that are missing all of the listed
                tags. (Works like OR logic).
                '''
        )
    parser_list_blocks.add_argument(
        "-T", "--must-have-tags", action="store",
        help='''Hash (#) separated list of tags (e.g. -t bash#proj-2#test_3).
                Filter out all code blocks that are missing at least one of the
                listed tags. (Works like AND logic).'''
        )
    parser_list_blocks.add_argument(
        "-N", "--must-not-have-tags", action="store",
        help='''Hash (#) separated list of tags (e.g. -t bash#proj-2#test_3).
                Filter out all code blocks that contain at least one of the
                listed tags. (Works like AND NOT logic).'''
        )
    parser_list_blocks.add_argument(
        "--light", action="store_true",
        help="Light terminal background."
        )
    parser_list_blocks.add_argument(
        "--pretty", action="store_true",
        help="Pretty print code blocks."
        )
    parser_list_blocks.set_defaults(
        tags="",
        must_have_tags="",
        must_not_have_tags="",
        pretty=False,
        )

    argcomplete.autocomplete(parser)
    return parser.parse_args()

def main():
    args = __parse_args()
    logging.basicConfig(
        format = '%(asctime)s.%(msecs)03d, %(levelname)s: %(message)s',
        datefmt = '%Y-%m-%d %H:%M:%S',
        filename = None,
        level = logging.DEBUG if args.debug else logging.WARNING,
        )
    if args.version:
        print("rundoc {} - Copyright {} {} <{}>".format(
            rundoc.__version__,
            rundoc.__year__,
            rundoc.__author__,
            rundoc.__author_email__,
            ))
        sys.exit(0)

    if args.cmd == 'run':
        try:
            commander = parsers.parse_doc(
                args.mkd_file,
                args.tags,
                args.must_have_tags,
                args.must_not_have_tags,
                darkbg=not args.light,
                )
        except BadEnv as e:
            print("{}{}{}".format(clr.red, e, clr.end))
            sys.exit(1)
        try:
            commander.run(
                step=args.step,
                yes=args.yes,
                inherit_env=args.inherit_env,
                pause=args.pause,
                retry=args.retry,
                retry_pause=args.retry_pause,
                output_file=args.output,
                )
        except KeyboardInterrupt:
            commander.die_with_grace()
            sys.exit(1)
        except BadEnv as e:
            print("{}{}{}".format(clr.red, e, clr.end))
            sys.exit(1)
        except CodeFailed as e:
            sys.exit(1)
    if args.cmd == 'replay':
        try:
            commander = parsers.parse_output(args.saved_output)
        except Exception as e:
            logger.error('Failed to parse file: {}'.format(e))
            sys.exit(1)
        try:
            commander.run(
                step=args.step,
                yes=True,
                inherit_env=False,
                pause=args.pause,
                retry=args.retry,
                retry_pause=args.retry_pause,
                output_file=args.output,
                )
        except KeyboardInterrupt:
            commander.die_with_grace()
            sys.exit(1)
        except BadEnv as e:
            print("{}{}{}".format(clr.red, e, clr.end))
            sys.exit(1)
        except CodeFailed as e:
            sys.exit(1)
    if args.cmd == 'list-tags':
        try:
            tags = parsers.get_tags(args.mkd_file)
            max_num_len = len(str(tags[0][0]))
            for key, value in tags:
                print("{}{}{}".format(
                    value, ' '*(max_num_len - len(str(value))), key))
        except Exception as e:
            logger.error('Failed to parse file: {}'.format(e))
            sys.exit(1)

    if args.cmd == 'list-blocks':
        try:
            parsers.print_blocks(
                args.mkd_file,
                args.tags,
                args.must_have_tags,
                args.must_not_have_tags,
                darkbg=not args.light,
                pretty=args.pretty,
                )
        except Exception as e:
            logger.error('Failed to parse file: {}'.format(e))
            sys.exit(1)

if __name__ == '__main__':
    main()

